@@ -1,7 +1,7 @@
 /* -*- indent-tabs-mode: t; tab-width: 8; c-basic-offset: 8; -*- */
 
 /* Copyright (c) 2004 - 2006 Derek Foreman, Ben Jansens
-   Copyright (c) 2006 - 2011 Thomas Schmitt <scdbackup@gmx.net>
+   Copyright (c) 2006 - 2012 Thomas Schmitt <scdbackup@gmx.net>
    Provided under GPL version 2 or later.
 */
 
@@ -73,6 +73,7 @@ int scsi_init_command(struct command *c, unsigned char *opcode, int oplen)
 	c->error = 0;
 	c->retry = 0;
 	c->page = NULL;
+	c->timeout = Libburn_scsi_default_timeouT;
 	return 1;
 }
 
@@ -101,7 +102,8 @@ int spc_decode_sense(unsigned char *sense, int senselen,
 }
 
 
-int spc_test_unit_ready_r(struct burn_drive *d, int *key, int *asc, int *ascq)
+int spc_test_unit_ready_r(struct burn_drive *d, int *key, int *asc, int *ascq,
+				int *progress)
 {
 	struct command *c;
 
@@ -114,8 +116,13 @@ int spc_test_unit_ready_r(struct burn_drive *d, int *key, int *asc, int *ascq)
 	c->dir = NO_TRANSFER;
 	d->issue_command(d, c);
 	*key = *asc = *ascq = 0;
+	*progress = -1;
 	if (c->error) {
 		spc_decode_sense(c->sense, 0, key, asc, ascq);
+		if (c->sense[0] == 0x70 &&
+		    ((c->sense[2] & 0x0f) == 0 || (c->sense[2] & 0x0f) == 2) &&
+		    (c->sense[15] & 0x80))
+			*progress = (c->sense[16] << 8) + c->sense[17];
 		return (key == 0);
 	}
 	return 1;
@@ -124,9 +131,9 @@ int spc_test_unit_ready_r(struct burn_drive *d, int *key, int *asc, int *ascq)
 
 int spc_test_unit_ready(struct burn_drive *d)
 {
-	int key,asc,ascq;
+	int key, asc, ascq, progress;
 
-	return spc_test_unit_ready_r(d, &key, &asc, &ascq);
+	return spc_test_unit_ready_r(d, &key, &asc, &ascq, &progress);
 }
 
 
@@ -140,7 +147,7 @@ int spc_wait_unit_attention(struct burn_drive *d, int max_sec, char *cmd_text,
 {
 	int i, ret = 1, key = 0, asc = 0, ascq = 0, clueless_start = 0;
 	static double tests_per_second = 2.0;
-	int sleep_usecs, loop_limit, clueless_timeout;
+	int sleep_usecs, loop_limit, clueless_timeout, progress;
 	char *msg = NULL;
 	unsigned char sense[14];
 
@@ -153,7 +160,7 @@ int spc_wait_unit_attention(struct burn_drive *d, int max_sec, char *cmd_text,
 		usleep(sleep_usecs);
 
 	for(i = !(flag & 1); i < loop_limit; i++) {
-		ret = spc_test_unit_ready_r(d, &key, &asc, &ascq);
+		ret = spc_test_unit_ready_r(d, &key, &asc, &ascq, &progress);
 		if (ret > 0) /* ready */
 	break;
 		if (key!=0x2 || asc!=0x4) {
@@ -253,17 +260,37 @@ void spc_request_sense(struct burn_drive *d, struct buffer *buf)
 	d->issue_command(d, c);
 }
 
+/* @return -2 = drive is ready , -1 = not ready, but no progress reported
+           >= 0 progress indication between 0 and 65535
+*/
 int spc_get_erase_progress(struct burn_drive *d)
 {
 	struct buffer *b = NULL;
-	int ret;
+	int ret, key, asc, ascq, progress;
 
 	if (mmc_function_spy(d, "get_erase_progress") <= 0)
 		{ret = 0; goto ex;}
 
+	/* ts B20104 :
+	   TEST UNIT READY seems to be more reliable than REQUEST SENSE.
+	   Nevertheless growisofs still uses the latter as fallback.
+	*/
+	ret = spc_test_unit_ready_r(d, &key, &asc, &ascq, &progress);
+	if (ret > 0)
+		{ret = -2; goto ex;}
+	if (progress >= 0)
+		{ret = progress; goto ex;}
+
+	/* Fallback to request sense */
 	BURN_ALLOC_MEM(b, struct buffer, 1);
 	spc_request_sense(d, b);
-	ret = (b->data[16] << 8) | b->data[17];
+
+	/* Checking the preconditions as of SPC-3 4.5.2.4.4 and 4.5.3 */
+	ret = -1;
+	if (b->data[0] == 0x70 &&
+	    ((b->data[2] & 0x0f) == 0 || (b->data[2] & 0x0f) == 2) &&
+	    (b->data[15] & 0x80))
+		ret = (b->data[16] << 8) | b->data[17];
 ex:;
 	BURN_FREE_MEM(b);
 	return ret;
@@ -718,8 +745,8 @@ Although command MODE SELECT is SPC, the content of the
 Write Parameters Mode Page (05h) is MMC (Table 108 in MMC-1). 
 Thus the filling of the mode page is done by mmc_compose_mode_page_5().
 */
-void spc_select_write_params(struct burn_drive *d,
-			     const struct burn_write_opts *o)
+void spc_select_write_params(struct burn_drive *d, struct burn_session *s,
+			     int tnum, const struct burn_write_opts *o)
 {
 	struct buffer *buf = NULL;
 	struct command *c = NULL;
@@ -776,7 +803,7 @@ void spc_select_write_params(struct burn_drive *d,
 	c->page->bytes = alloc_len;
 
 	/* ts A61229 */
-	if (mmc_compose_mode_page_5(d, o, c->page->data + 8) <= 0)
+	if (mmc_compose_mode_page_5(d, s, tnum, o, c->page->data + 8) <= 0)
 		goto ex;
 
 	c->dir = TO_DRIVE;
@@ -1041,6 +1068,11 @@ enum response scsi_error_msg(struct burn_drive *d, unsigned char *sense,
 	sprintf(msg, "[%X %2.2X %2.2X] ", *key, *asc, *ascq);
 	msg= msg + strlen(msg);
 
+	if (key_def[*key & 0xf][0] != '(') {
+		sprintf(msg, "%s. ", key_def[*key & 0xf]);
+		msg= msg + strlen(msg);
+	}
+
 	switch (*asc) {
 	case 0x00:
 		if (*key > 0 || *ascq > 0) 
@@ -1050,17 +1082,15 @@ enum response scsi_error_msg(struct burn_drive *d, unsigned char *sense,
 
 	case 0x02:
 		sprintf(msg, "Not ready");
-		return RETRY;
+		goto return_retry;
 	case 0x04:
 		if (*ascq == 1)
 			sprintf(msg,
 			"Logical unit is in the process of becoming ready");
 		else
 			sprintf(msg, "Logical unit is not ready");
-		return RETRY;
+		goto return_retry;
 	case 0x08:
-		if (*key != 4)
-			break;
 		if (*ascq == 0)
 			sprintf(msg, "Logical unit communication failure");
 		else if (*ascq == 1)
@@ -1069,10 +1099,10 @@ enum response scsi_error_msg(struct burn_drive *d, unsigned char *sense,
 			sprintf(msg, "Logical unit communication parity error");
 		else if (*ascq == 3)
 			sprintf(msg, "Logical unit communication crc error");
-		return RETRY;
-	case 0x09:
-		if (*key != 4)
+		else
 			break;
+		goto return_retry;
+	case 0x09:
 		if (*ascq == 0)
 			sprintf(msg, "Track following error");
 		else if (*ascq == 1)
@@ -1085,24 +1115,27 @@ enum response scsi_error_msg(struct burn_drive *d, unsigned char *sense,
 			sprintf(msg, "Head select fault");
 		else
 			break;
-		return FAIL;
+		goto return_fail;
 	case 0x0C:
-		if (*key == 2 && *ascq == 7)
-			sprintf(msg, "Write error, recovery needed"); 
-		else if (*key == 2 && *ascq == 0x0f)
-			sprintf(msg, "Defects in error window"); 
-		else if (*key == 3 && *ascq == 2)
-			sprintf(msg, "Write error, auto reallocation failed");
-		else if (*key == 3 && *ascq == 9)
-			sprintf(msg, "Write error, loss of streaming");
-		else if (*key == 3)
+		if (*ascq == 0)
 			sprintf(msg, "Write error");
+		else if (*ascq == 1)
+			sprintf(msg,
+				"Write error, recovered with auto-allocation");
+		else if (*ascq == 2)
+			sprintf(msg, "Write error, auto reallocation failed");
+		else if (*ascq == 7)
+			sprintf(msg, "Write error, recovery needed"); 
+		else if (*ascq == 8)
+			sprintf(msg, "Write error, recovery failed"); 
+		else if (*ascq == 9)
+			sprintf(msg, "Write error, loss of streaming");
+		else if (*ascq == 0x0f)
+			sprintf(msg, "Defects in error window"); 
 		else
 			break;
-		return FAIL;
+		goto return_fail;
 	case 0x11:
-		if (*key != 3)
-			break;
 		if (*ascq == 0)
 			sprintf(msg, "Unrecovered read error");
 		else if (*ascq == 1)
@@ -1115,70 +1148,73 @@ enum response scsi_error_msg(struct burn_drive *d, unsigned char *sense,
 			sprintf(msg, "CIRC uncorrectable error");
 		else
 			break;
-		return FAIL;
+		goto return_fail;
 	case 0x15:
-		if (*key != 3 && *key != 4)
+		if (*ascq == 0)
+			sprintf(msg, "Random positioning error");
+		else if (*ascq == 1)
+			sprintf(msg, "Mechanical positioning error");
+		else
 			break;
-		sprintf(msg, "Random positioning error");
-		return FAIL;
+		goto return_fail;
 	case 0x1a:
-		if (*key != 5)
+		if (*ascq == 0)
+			sprintf(msg, "Parameter list length error");
+		else
 			break;
-		sprintf(msg, "Parameter list length error");
-		return FAIL;
+		goto return_fail;
 	case 0x1b:
-		if (*key != 4)
+		if (*ascq == 0)
+			sprintf(msg, "Synchronous data transfer error");
+		else
 			break;
-		sprintf(msg, "Synchronous data transfer error");
-		return FAIL;
+		goto return_fail;
 	case 0x20:
-		if (*key != 5)
+		if (*ascq == 0)
+			sprintf(msg, "Invalid command operation code");
+		else
 			break;
-		sprintf(msg, "Invalid command operation code");
-		return FAIL;
+		goto return_fail;
 	case 0x21:
-		if (*key != 5)
-			break;
 		if (*ascq == 0)
 			sprintf(msg, "Lba out of range");
+		else if (*ascq == 1)
+			sprintf(msg, "Invalid element address");
+		else if (*ascq == 2)
+			sprintf(msg, "Invalid address for write");
 		else if (*ascq == 3)
 			sprintf(msg, "Invalid write crossing layer jump");
 		else
-			sprintf(msg, "Invalid address");
-		return FAIL;
+			break;
+		goto return_fail;
 	case 0x24:
-		if (*key != 5)
+		if (*ascq == 0)
+			sprintf(msg, "Invalid field in cdb");
+		else
 			break;
-		sprintf(msg, "Invalid field in cdb");
-		return FAIL;
+		goto return_fail;
 	case 0x26:
-		if (*key != 5)
-			break;
-		if (*ascq == 1)
+		if (*ascq == 0)
+			sprintf(msg, "Invalid field in parameter list");
+		else if (*ascq == 1)
 			sprintf(msg, "Parameter not supported");
 		else if (*ascq == 2)
 			sprintf(msg, "Parameter value invalid");
 		else
-			sprintf(msg, "Invalid field in parameter list");
-		return FAIL;
+			break;
+		goto return_fail;
 	case 0x27:
-		if (*key != 7)
-			break;
 		sprintf(msg, "Write protected");
-		return FAIL;
+		goto return_fail;
 	case 0x28:
-		if (*key != 6)
-			break;
 		if (*ascq == 0)
 			sprintf(msg, "Medium may have changed");
 		else if (*ascq == 2)
 			sprintf(msg, "Format layer may have changed");
 		else
 			break;
-		return RETRY;
+		goto return_retry;
 	case 0x29:
-		if (*key != 6)
-			break;
 		if (*ascq == 0)
 			sprintf(msg,
                                "Power on, reset, or bus device reset occured");
@@ -1192,28 +1228,23 @@ enum response scsi_error_msg(struct burn_drive *d, unsigned char *sense,
 			sprintf(msg, "Device internal reset");
 		else
 			break;
-		return RETRY;
+		goto return_retry;
 	case 0x2c:
-		if (*key != 5)
-			break;
 		if (*ascq == 0)
 			sprintf(msg, "Command sequence error");
 		else 
 			break;
-		return FAIL;
+		goto return_fail;
 	case 0x2e:
-		if (*key != 6)
-			break;
 		if (*ascq == 0)
-			sprintf(msg,
-                               "Insufficient time for operation");
+			sprintf(msg, "Insufficient time for operation");
 		else 
 			break;
-		return FAIL;
+		goto return_fail;
 	case 0x30:
-		if (*key != 2 && *key != 5)
-			break;
-		if (*ascq == 1)
+		if (*ascq == 0)
+			sprintf(msg, "Incompatible medium installed");
+		else if (*ascq == 1)
 			sprintf(msg, "Cannot read medium, unknown format");
 		else if (*ascq == 2)
 			sprintf(msg,
@@ -1229,47 +1260,50 @@ enum response scsi_error_msg(struct burn_drive *d, unsigned char *sense,
 		else if (*ascq == 7)
 			sprintf(msg, "Cleaning failure");
 		else
-			sprintf(msg, "Incompatible medium installed");
-		return FAIL;
-	case 0x31:
-		if (*key != 3)
 			break;
+		goto return_fail;
+	case 0x31:
 		if (*ascq == 0)
 			sprintf(msg, "Medium unformatted or format corrupted");
 		else if (*ascq == 1)
 			sprintf(msg, "Format command failed");
-		return FAIL;
-	case 0x3A:
-		if (*key != 2)
+		else
 			break;
-		if (*ascq == 1)
+		goto return_fail;
+	case 0x3A:
+		if (*ascq == 0)
+			sprintf(msg, "Medium not present");
+		else if (*ascq == 1)
 			sprintf(msg, "Medium not present, tray closed");
 		else if (*ascq == 2)
 			sprintf(msg, "Medium not present, tray open");
 		else if (*ascq == 3)
 			sprintf(msg, "Medium not present, loadable");
 		else
-			sprintf(msg, "Medium not present");
+			break;
 		d->status = BURN_DISC_EMPTY;
-		return FAIL;
+		goto return_fail;
 	case 0x3E:
 		if (*ascq == 1)
 			sprintf(msg, "Logical unit failure");
 		else if (*ascq == 2)
 			sprintf(msg, "Timeout on logical unit");
-		return FAIL;
+		else
+			break;
+		goto return_fail;
 	case 0x44:
 		if (*ascq == 0)
 			sprintf(msg, "Internal target failure");
-		return FAIL;
+		else
+			break;
+		goto return_fail;
 	case 0x57:
-		if (*key != 3 || *ascq != 0)
+		if (*ascq == 0)
+			sprintf(msg, "Unable to recover Table-of-Content");
+		else
 			break;
-		sprintf(msg, "Unable to recover Table-of-Content");
-		return FAIL;
+		goto return_fail;
 	case 0x63:
-		if (*key != 5)
-			break;
 		if (*ascq == 0)
 			sprintf(msg,
 				"End of user area encountered on this track");
@@ -1277,51 +1311,67 @@ enum response scsi_error_msg(struct burn_drive *d, unsigned char *sense,
 			sprintf(msg, "Packet does not fit in available space");
 		else
 			break;
-		return FAIL;
+		goto return_fail;
 	case 0x64:
-		if (*key != 5)
-			break;
 		if (*ascq == 0)
 			sprintf(msg, "Illegal mode for this track");
 		else if (*ascq == 1)
 			sprintf(msg, "Invalid packet size");
 		else
 			break;
-		return FAIL;
+		goto return_fail;
 	case 0x72:
-		if (*key == 3)
+		if (*ascq == 0)
 			sprintf(msg, "Session fixation error");
-		else if (*key == 5 && *ascq == 3)
+		else if (*ascq == 1)
+			sprintf(msg, "Session fixation error writing lead-in");
+		else if (*ascq == 2)
+			sprintf(msg,
+				"Session fixation error writing lead-out");
+		else if (*ascq == 3)
 			sprintf(msg,
 			"Session fixation error, incomplete track in session");
-		else if (*key == 5 && *ascq == 4)
+		else if (*ascq == 4)
 			sprintf(msg,
-			"Empty or partially written reserved track");
-		else if (*key == 5 && *ascq == 5)
-			sprintf(msg,
-			"No more track reservations allowed");
+				"Empty or partially written reserved track");
+		else if (*ascq == 5)
+			sprintf(msg, "No more track reservations allowed");
 		else
 			break;
-		return FAIL;
+		goto return_fail;
 	case 0x73:
-		if (*key == 3 && *ascq == 0)
+		if (*ascq == 0)
 			sprintf(msg, "CD control error");
-		else if (*key == 3 && *ascq == 2)
+		else if (*ascq == 1)
+			sprintf(msg, "Power calibration area almost full");
+		else if (*ascq == 2)
 			sprintf(msg, "Power calibration area is full");
-		else if (*key == 3 && *ascq == 3)
+		else if (*ascq == 3)
 			sprintf(msg, "Power calibration area error");
-		else if (*key == 3 && *ascq == 4)
+		else if (*ascq == 4)
 			sprintf(msg, "Program memory area update failure");
-		else if (*key == 3 && *ascq == 5)
+		else if (*ascq == 5)
 			sprintf(msg, "Program memory area is full");
 		else
 			break;
-		return FAIL;
+		goto return_fail;
 	}
 	sprintf(msg_data,
 		"See MMC specs: Sense Key %X \"%s\", ASC %2.2X ASCQ %2.2X",
 		*key & 0xf, key_def[(*key) & 0xf], *asc, *ascq);
+	goto return_fail;
+
+return_fail:
+	strcat(msg, ".");
+	if (*key == 1)
+		return GO_ON;
 	return FAIL;
+
+return_retry:;
+	strcat(msg, ".");
+	if (*key == 1)
+		return GO_ON;
+	return RETRY;
 }
 
 
@@ -1655,38 +1705,64 @@ int scsi_eval_cmd_outcome(struct burn_drive *d, struct command *c, void *fp,
 			int loop_count, int flag)
 {
 	enum response outcome;
-	int done = -1, usleep_time;
+	int done = -1, usleep_time, ret;
+	char *msg = NULL;
 
 	if (burn_sg_log_scsi & 3)
 		scsi_log_err(c, fp, sense, sense_len, duration,
 				 (sense_len > 0) | (flag & 2));
 	if (sense_len <= 0)
-		return 1;
+		{done = 1; goto ex;}
 		
 	outcome = scsi_error(d, sense, sense_len);
 	if (outcome == RETRY && c->retry && !(flag & 1)) {
 		/* Calming down retries and breaking up endless cycle
 		*/
-		usleep_time = Libburn_scsi_retry_usleeP +
-				loop_count * Libburn_scsi_retry_incR;
+		if (c->opcode[0] == 0x2A || c->opcode[0] == 0xAA) {
+			/* WRITE(10) , WRITE(12) */
+			usleep_time = Libburn_scsi_write_retry_usleeP +
+				loop_count * Libburn_scsi_write_retry_incR;
+			if (usleep_time > Libburn_scsi_write_retry_umaX)
+				usleep_time = Libburn_scsi_write_retry_umaX;
+		} else {
+			usleep_time = Libburn_scsi_retry_usleeP +
+					loop_count * Libburn_scsi_retry_incR;
+			if (usleep_time > Libburn_scsi_retry_umaX)
+				usleep_time = Libburn_scsi_retry_umaX;
+		}
 		if (time(NULL) + usleep_time / 1000000 - start_time >
 		    timeout_ms / 1000 + 1) {
+			BURN_ALLOC_MEM(msg, char, 320);
+			sprintf(msg,
+				"Timeout exceed (%d ms). Retry canceled.\n",
+				timeout_ms);
+			libdax_msgs_submit(libdax_messenger, d->global_index,
+				0x0002018a,
+				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				msg, 0, 0);
 			done = 1;
-			goto ex;
+			goto err_ex;
 		}
-		usleep(usleep_time);
+		if (d->cancel)
+			{done = 1; goto ex;}
+		if (usleep_time > 0)
+			usleep(usleep_time);
+		if (d->cancel)
+			{done = 1; goto ex;}
 		if (burn_sg_log_scsi & 3) 
 			scsi_log_cmd(c, fp, 0);
-		return 0;
+		{done = 0; goto ex;}
 	} else if (outcome == RETRY) {
 		done = 1;
 	} else if (outcome == GO_ON) {
-		return 1;
+		{done = 1; goto ex;}
 	} else if (outcome == FAIL) {
 		done = 1;
 	}
-ex:;
+err_ex:;
 	c->error = 1;
 	scsi_notify_error(d, c, sense, sense_len, 0);
+ex:;
+	BURN_FREE_MEM(msg);
 	return done;
 }
