@@ -1,7 +1,7 @@
 /* -*- indent-tabs-mode: t; tab-width: 8; c-basic-offset: 8; -*- */
 
 /* Copyright (c) 2004 - 2006 Derek Foreman, Ben Jansens
-   Copyright (c) 2006 - 2011 Thomas Schmitt <scdbackup@gmx.net>
+   Copyright (c) 2006 - 2012 Thomas Schmitt <scdbackup@gmx.net>
    Provided under GPL version 2 or later.
 */
 
@@ -272,7 +272,6 @@ int burn_write_close_track(struct burn_write_opts *o, struct burn_session *s,
 }
 
 
-
 /* ts A61030 */
 int burn_write_close_session(struct burn_write_opts *o)
 {
@@ -298,49 +297,74 @@ int burn_write_close_session(struct burn_write_opts *o)
 }
 
 
-/* ts A60819:
-   This is unused since about Feb 2006, icculus.org/burn CVS.
-   The compiler complains. We shall please our compiler.
+/* ts A60819, B20101:
+   This is useful only when changes about CD SAO get tested.
+ # define Libburn_write_with_function_print_cuE yes
 */
+
 #ifdef Libburn_write_with_function_print_cuE
 
+
+static char cue_printify(char c)
+{
+	if (c >= 32 && c < 127)
+		return c;
+	return '#';
+}
+
+
 static void print_cue(struct cue_sheet *sheet)
 {
 	int i;
 	unsigned char *unit;
 
 	printf("\n");
-	printf("ctladr|trno|indx|form|scms|  msf\n");
-	printf("------+----+----+----+----+--------\n");
+	printf("ctladr|trno|indx|form|scms|   msf    |  text\n");
+	printf("------+----+----+----+----+----------+--------\n");
 	for (i = 0; i < sheet->count; i++) {
 		unit = sheet->data + 8 * i;
-		printf(" %1X  %1X | %02X | %02X | %02X | %02X |",
-		       (unit[0] & 0xf0) >> 4, unit[0] & 0xf, unit[1], unit[2],
-		       unit[3], unit[4]);
-		printf("%02X:%02X:%02X\n", unit[5], unit[6], unit[7]);
+		if ((unit[0] & 0xf) == 2) {
+			printf(
+		" %1X  %1X |    |    |    |    |          | %c%c%c%c%c%c%c\n",
+				(unit[0] & 0xf0) >> 4, unit[0] & 0xf,
+				cue_printify(unit[1]), cue_printify(unit[2]), 
+				cue_printify(unit[3]), cue_printify(unit[4]), 
+				cue_printify(unit[5]), cue_printify(unit[6]), 
+				unit[7] == 0 ? ' ' : cue_printify(unit[7]));
+		} else if ((unit[0] & 0xf) == 3) {
+			printf(
+		" %1X  %1X | %2d |    |    |    |          | %c%c%c%c%c%c\n",
+				(unit[0] & 0xf0) >> 4, unit[0] & 0xf,
+				unit[1], cue_printify(unit[2]), 
+				cue_printify(unit[3]), cue_printify(unit[4]), 
+				cue_printify(unit[5]), cue_printify(unit[6]), 
+				cue_printify(unit[7]));
+		} else if (unit[1] > 99) {
+			printf(" %1X  %1X |0x%02X| %2d | %02X | %02X |",
+				(unit[0] & 0xf0) >> 4, unit[0] & 0xf,
+				unit[1], unit[2], unit[3], unit[4]);
+			printf(" %02d:%02d:%02d |\n",
+				unit[5], unit[6], unit[7]);
+		} else {
+			printf(" %1X  %1X | %2d | %2d | %02X | %02X |",
+				(unit[0] & 0xf0) >> 4, unit[0] & 0xf,
+				unit[1], unit[2], unit[3], unit[4]);
+			printf(" %02d:%02d:%02d |\n",
+				unit[5], unit[6], unit[7]);
+		}
 	}
+	fflush(stdout);
 }
 
 #endif /* Libburn_write_with_print_cuE */
 
 
-/* ts A61009 : changed type from void to int */
-/** @return 1 = success , <=0 failure */
-static int add_cue(struct cue_sheet *sheet, unsigned char ctladr,
-		    unsigned char tno, unsigned char indx,
-		    unsigned char form, unsigned char scms, int lba)
+/* ts B11226 */
+static int new_cue(struct cue_sheet *sheet, int number, int flag)
 {
-	unsigned char *unit;
 	unsigned char *ptr;
-	int m, s, f;
 
-	burn_lba_to_msf(lba, &m, &s, &f);
-
-	sheet->count++;
-	ptr = realloc(sheet->data, sheet->count * 8);
-
-	/* ts A61009 */
-	/* a ssert(ptr); */
+	ptr = realloc(sheet->data, (sheet->count + number) * 8);
 	if (ptr == NULL) {
 		libdax_msgs_submit(libdax_messenger, -1, 0x00020111,
 			LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
@@ -348,8 +372,26 @@ static int add_cue(struct cue_sheet *sheet, unsigned char ctladr,
 			0, 0);
 		return -1;
 	}
-
 	sheet->data = ptr;
+	sheet->count += number;
+	return 1;
+}
+
+
+/* ts B11226 : outsourced new_cue() */
+/** @return 1 = success , <=0 failure */
+static int add_cue(struct cue_sheet *sheet, unsigned char ctladr,
+		    unsigned char tno, unsigned char indx,
+		    unsigned char form, unsigned char scms, int lba)
+{
+	unsigned char *unit;
+	int m, s, f, ret;
+
+	burn_lba_to_msf(lba, &m, &s, &f);
+
+	ret = new_cue(sheet, 1, 0);
+	if (ret <= 0)
+		return -1;
 	unit = sheet->data + (sheet->count - 1) * 8;
 	unit[0] = ctladr;
 	unit[1] = tno;
@@ -362,13 +404,60 @@ static int add_cue(struct cue_sheet *sheet, unsigned char ctladr,
 	return 1;
 }
 
+
+/* ts B11226 */
+static int add_catalog_cue(struct cue_sheet *sheet, unsigned char catalog[13])
+{
+	unsigned char *unit;
+	int i, ret;
+
+	ret = new_cue(sheet, 2, 0);
+	if (ret <= 0)
+		return -1;
+	unit = sheet->data + (sheet->count - 2) * 8;
+	unit[0] = unit[8] = 0x02;
+	for (i = 0; i < 13; i++)
+		unit[1 + (i >= 7) * 8 + (i % 7)] = catalog[i];
+	unit[15] = 0x00;
+	return 1;
+}
+
+
+/* ts B11226 */
+static int add_isrc_cue(struct cue_sheet *sheet, unsigned char ctladr, int tno,
+			struct isrc *isrc)
+{
+	unsigned char *unit;
+	int i, ret;
+	char text[8];
+
+	ret = new_cue(sheet, 2, 0);
+	if (ret <= 0)
+		return -1;
+	unit = sheet->data + (sheet->count - 2) * 8;
+	unit[0] = unit[8] = (ctladr & 0xf0) | 0x03;
+	unit[1] = unit[9] = tno;
+	unit[2] = isrc->country[0];
+	unit[3] = isrc->country[1];
+	unit[4] = isrc->owner[0];
+	unit[5] = isrc->owner[1];
+	unit[6] = isrc->owner[2];
+	sprintf(text, "%-2.2u%-5.5u", (unsigned int) isrc->year, isrc->serial);
+	unit[7] = text[0];
+	for (i = 1; i < 7; i++)
+		unit[9 + i] = text[i];
+	return 1;
+}
+
+
 /* ts A61114: added parameter nwa */
 struct cue_sheet *burn_create_toc_entries(struct burn_write_opts *o,
 					  struct burn_session *session,
 					  int nwa)
 {
 	int i, m, s, f, form, pform, runtime = -150, ret, track_length;
-	unsigned char ctladr;
+	int leadin_form, leadin_start, pregap = 150, postgap;
+	unsigned char ctladr, scms;
 	struct burn_drive *d;
 	struct burn_toc_entry *e;
 	struct cue_sheet *sheet;
@@ -376,6 +465,19 @@ struct cue_sheet *burn_create_toc_entries(struct burn_write_opts *o,
 	int ntr = session->tracks;
 	int rem = 0;
 
+#define Libburn_track_multi_indeX yes
+
+#ifdef Libburn_track_multi_indeX
+	int j;
+#endif
+
+	if (ntr < 1) {
+		libdax_msgs_submit(libdax_messenger, -1, 0x0002019c,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			"Session has no defined tracks", 0, 0);
+		return NULL;
+	}
+
 	d = o->drive;
 
 #ifdef Libburn_sao_can_appenD
@@ -404,13 +506,46 @@ struct cue_sheet *burn_create_toc_entries(struct burn_write_opts *o,
 			"Track mode has unusable value", 0, 0);
 		goto failed;
 	}
-	ret = add_cue(sheet, ctladr | 1, 0, 0, 1, 0, runtime);
+	if (o->num_text_packs > 0) {
+		leadin_form = 0x41;
+	} else {
+		leadin_form = 0x01;
+
+		/* Check for CD-TEXT in session. Not the final creation,
+		   because the cue sheet content might be needed for CD-TEXT
+		   pack type 0x88 "TOC".
+		 */
+		if (o->text_packs == NULL) {
+			ret = burn_cdtext_from_session(session, NULL, NULL, 1);
+			if (ret < 0)
+				goto failed;
+			else if (ret > 0)
+				leadin_form = 0x41;
+		}
+	}
+
+	if (o->has_mediacatalog)
+		ret = add_catalog_cue(sheet, o->mediacatalog);
+	else if (session->mediacatalog[0])
+		ret = add_catalog_cue(sheet, session->mediacatalog);
+	else
+		ret = 1;
 	if (ret <= 0)
 		goto failed;
-	ret = add_cue(sheet, ctladr | 1, 1, 0, form, 0, runtime);
+
+	/* ts B11225 
+	   MMC-5 6.33.3.15 Data Form of Sub-channel
+	   seems to indicate that for leadin_form 0x41 one should announce
+	   d->start_lba as start of the leadin (e.g. -12490) and that data
+	   block type should 2 or 3 with mode page 05h. But my drives refuse
+	   on that.
+           It works with LBA -150 and data block type 0. Shrug.
+	*/
+	leadin_start = runtime;
+	ret = add_cue(sheet, (ctladr & 64) | 1, 0, 0, leadin_form, 0,
+								leadin_start);
 	if (ret <= 0)
 		goto failed;
-	runtime += 150;
 
 	d->toc_entries = ntr + 3;
 
@@ -432,6 +567,13 @@ struct cue_sheet *burn_create_toc_entries(struct burn_write_opts *o,
 		goto failed;
 		*/
 	}
+	if (session->firsttrack + ntr - 1 > 99) {
+		libdax_msgs_submit(libdax_messenger, -1, 0x0002019b,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			"CD track number exceeds 99", 0, 0);
+		goto failed;
+	}
+	session->lasttrack = session->firsttrack + ntr - 1;
 
 	d->toc_entry = calloc(d->toc_entries, sizeof(struct burn_toc_entry));
 	e = d->toc_entry;
@@ -440,11 +582,11 @@ struct cue_sheet *burn_create_toc_entries(struct burn_write_opts *o,
 		e[0].control = TOC_CONTROL_AUDIO;
 	else
 		e[0].control = TOC_CONTROL_DATA;
-	e[0].pmin = 1;
+	e[0].pmin = session->firsttrack;
 	e[0].psec = o->format;
 	e[0].adr = 1;
 	e[1].point = 0xA1;
-	e[1].pmin = ntr;
+	e[1].pmin = session->lasttrack;
 	e[1].adr = 1;
 	if (tar[ntr - 1]->mode & BURN_AUDIO)
 		e[1].control = TOC_CONTROL_AUDIO;
@@ -454,26 +596,119 @@ struct cue_sheet *burn_create_toc_entries(struct burn_write_opts *o,
 	e[2].control = e[1].control;
 	e[2].adr = 1;
 
-	/* ts A70121 : The pause before the first track is not a Pre-gap.
-	   To count it as part 2 of a Pre-gap is a dirty hack. It also seems
-	   to have caused confusion in dealing with part 1 of an eventual
-	   real Pre-gap. mmc5r03c.pdf 6.33.3.2, 6.33.3.18 .
-	*/
 	tar[0]->pregap2 = 1;
+	if (tar[0]->pregap2_size < 150)
+		tar[0]->pregap2_size = 150;
 
 	pform = form;
 	for (i = 0; i < ntr; i++) {
-		type_to_form(tar[i]->mode, &ctladr, &form);
 
+		/* ts A70125 : 
+		   Still not understanding the sense behind linking tracks,
+		   i decided to at least enforce the MMC specs' minimum
+		   track length.
+		*/ 
+		track_length = burn_track_get_sectors_2(tar[i], 1);
+		if (track_length < 300 && !burn_track_is_open_ended(tar[i])) {
+			track_length = 300;
+			if (!tar[i]->pad)
+				tar[i]->pad = 1;
+			burn_track_set_sectors(tar[i], track_length);
+		}
+
+		type_to_form(tar[i]->mode, &ctladr, &form);
+		if (tar[i]->mode & BURN_SCMS)
+			scms = 0x80;
+		else
+			scms = 0; 
+
+		if (tar[i]->isrc.has_isrc) {
+			ret = add_isrc_cue(sheet, ctladr,
+				i + session->firsttrack, &(tar[i]->isrc));
+			if (ret <= 0)
+				goto failed;
+		}
+		pregap = 0;
+		if (tar[i]->pregap2)
+			pregap = tar[i]->pregap2_size;
+		postgap = 0;
+		if (tar[i]->postgap) {
+			if (tar[i]->indices >= 99) {
+				libdax_msgs_submit(libdax_messenger, -1,
+					0x0002019a, LIBDAX_MSGS_SEV_SORRY,
+					LIBDAX_MSGS_PRIO_HIGH,
+					"Post-gap index number exceeds 99",
+					0, 0);
+				goto failed;
+			}
+			if (tar[i]->indices < 2)
+				tar[i]->indices = 2;
+			tar[i]->index[tar[i]->indices] = track_length;
+			postgap = tar[i]->postgap_size;
+		}
+
+#ifdef Libburn_track_multi_indeX
+
+		for(j = 0; j < (tar[i]->indices + !!tar[i]->postgap) || j < 2;
+									j++) {
+			if(tar[i]->index[j] == 0x7fffffff) {
+				if (j > 1)
+		break;
+				if (j == 0 && pregap <= 0)
+		continue;
+				/* force existence of mandatory index */
+				tar[i]->index[j] = 0;
+			} else if (j == 0) {
+				tar[i]->index[j] = 0;
+			} else if (j == 1 && tar[i]->index[0] == 0x7fffffff) {
+				tar[i]->index[j] = 0;
+			}
+
+			if (j == 1) {
+				tar[i]->entry = &e[3 + i];
+				e[3 + i].point = i + session->firsttrack;
+				burn_lba_to_msf(runtime, &m, &s, &f);
+				e[3 + i].pmin = m;
+				e[3 + i].psec = s;
+				e[3 + i].pframe = f;
+				e[3 + i].adr = 1;
+				e[3 + i].control = type_to_ctrl(tar[i]->mode);
+			}
+
+			/* >>> ??? else if j == 0 && mode change to -data :
+							Extended pregap */;
+
+			/* >>> check index with track size */;
+
+			tar[i]->index[j] += runtime;
+			ret = add_cue(sheet, ctladr | 1,
+					i + session->firsttrack, j, form, scms,
+					tar[i]->index[j]);
+			if (ret <= 0)
+				goto failed;
+			runtime += pregap;
+			pregap = 0;
+		}
+
+		runtime += track_length + postgap;
+
+#else /* Libburn_track_multi_indeX */
+
+		if (i == 0) {
+			ret = add_cue(sheet, ctladr | 1, session->firsttrack,
+					0, form, 0, runtime);
+			if (ret <= 0)
+				goto failed;
+			runtime += 150;
+		} else if (pform != form) {
 
 		/* ts A70121 : This seems to be thw wrong test. Correct would
 		   be to compare tar[]->mode or bit2 of ctladr.
 		*/ 
 
-		if (pform != form) {
-
-			ret = add_cue(sheet, ctladr | 1, i + 1, 0, form, 0,
-					 runtime);
+			ret = add_cue(sheet, ctladr | 1,
+					i + session->firsttrack, 0, form, scms,
+					runtime);
 			if (ret <= 0)
 				goto failed;
 
@@ -494,6 +729,10 @@ struct cue_sheet *burn_create_toc_entries(struct burn_write_opts *o,
 
    Next one has to care for Post-gap: table 555 in mmc5r03c.pdf does not
    show any although 6.33.3.19 would prescribe some.
+   ts B20111: Table 1 of MMC-1 shows two post-gaps. The first matches the
+              precriptions with SEND CUE SHEET. The second one is riddling.
+              Both are part of a track and occupy the range of the last index
+              of the track. Length is 2 seconds for each.
 
    Nobody seems to have ever tested this situation, up to now.
    It is banned for now in burn_disc_write().
@@ -505,7 +744,7 @@ struct cue_sheet *burn_create_toc_entries(struct burn_write_opts *o,
 /* XXX HERE IS WHERE WE DO INDICES IN THE CUE SHEET */
 /* XXX and we should make sure the gaps conform to ecma-130... */
 		tar[i]->entry = &e[3 + i];
-		e[3 + i].point = i + 1;
+		e[3 + i].point = i + session->firsttrack;
 		burn_lba_to_msf(runtime, &m, &s, &f);
 		e[3 + i].pmin = m;
 		e[3 + i].psec = s;
@@ -513,24 +752,16 @@ struct cue_sheet *burn_create_toc_entries(struct burn_write_opts *o,
 		e[3 + i].adr = 1;
 		e[3 + i].control = type_to_ctrl(tar[i]->mode);
 
-		ret = add_cue(sheet, ctladr | 1, i + 1, 1, form, 0, runtime);
+		ret = add_cue(sheet, ctladr | 1, i + session->firsttrack,
+				1, form, scms, runtime);
 		if (ret <= 0)
 			goto failed;
 
-		/* ts A70125 : 
-		   Still not understanding the sense behind linking tracks,
-		   i decided to at least enforce the MMC specs' minimum
-		   track length.
-		*/ 
-		track_length = burn_track_get_sectors(tar[i]);
-		if (track_length < 300 && !burn_track_is_open_ended(tar[i])) {
-			track_length = 300;
-			if (!tar[i]->pad)
-				tar[i]->pad = 1;
-			burn_track_set_sectors(tar[i], track_length);
-		}
 		runtime += track_length;
 
+#endif /* ! Libburn_track_multi_indeX */
+
+
 /* if we're padding, we'll clear any current shortage.
    if we're not, we'll slip toc entries by a sector every time our
    shortage is more than a sector
@@ -650,11 +881,155 @@ int burn_write_leadout(struct burn_write_opts *o,
 	return 1;
 }
 
+
+static int burn_create_text_packs(struct burn_write_opts *o,
+					struct burn_session *s,
+					int flag)
+{
+	int ret, num_packs = 0;
+	unsigned char *text_packs = NULL;
+
+	ret = burn_cdtext_from_session(s, &text_packs, &num_packs, 0);
+	if (ret > 0) {
+		if (o->text_packs != NULL)
+			free(o->text_packs);
+		o->text_packs = text_packs;
+		o->num_text_packs = num_packs;
+	}
+	return(ret);
+}
+
+
+static int burn_write_leadin_cdtext(struct burn_write_opts *o,
+					 struct burn_session *s, int flag)
+{
+	int ret, i, j, si, lba, sub_cursor = 0, err, write_lba, sectors = 0;
+	int self_made_text_packs = 0;
+	unsigned char *subdata = NULL;
+	struct burn_drive *d = o->drive;
+	struct buffer *buf = NULL;
+	enum burn_drive_status was_busy = o->drive->busy;
+#ifdef Libburn_debug_cd_texT
+	unsigned char *packs;
+#endif
+
+	if (o->num_text_packs <= 0) {
+		if (o->text_packs != NULL)
+			{ret = 1; goto ex;}
+		/* Try to create CD-TEXT from .cdtext_* of session and track */
+		ret = burn_create_text_packs(o, s, 0);
+		self_made_text_packs = 1;
+		if (ret <= 0)
+			goto ex;
+		if (o->num_text_packs <= 0)
+			{ret = 1; goto ex;}
+	}
+
+	if (!o->no_text_pack_crc_check) {
+		ret = burn_cdtext_crc_mismatches(o->text_packs,
+				 o->num_text_packs, 0);
+		if (ret != 0) {
+			libdax_msgs_submit(libdax_messenger, -1, 0x0002018f,
+				LIBDAX_MSGS_SEV_FAILURE, LIBDAX_MSGS_PRIO_HIGH,
+				"Program error: CD-TEXT pack CRC mismatch",
+				0, 0);
+			{ ret = 0; goto ex; }
+		}
+	}
+
+	d->busy = BURN_DRIVE_WRITING_LEADIN;
+
+#ifdef Libburn_debug_cd_texT
+	packs = o->text_packs;
+	fprintf(stderr,
+		"libburn_DEBUG: 8 bit CD-TEXT packs to be transmitted:\n"); 
+	for (i = 0; i < 18 * o->num_text_packs; i += 18) {
+		fprintf(stderr, "%4d :", i / 18);
+		for (j = 0; j < 18; j++) {
+			if (j >= 4 && j <= 15 && packs[i + j] >= 32 &&
+			    packs[i + j] <= 126 && packs[i] != 0x88 &&
+			    packs[i] != 0x89 && packs[i] != 0x8f)
+				fprintf(stderr, "  %c", packs[i + j]);
+			else
+				fprintf(stderr, " %2.2X", packs[i + j]);
+		}
+		fprintf(stderr, "\n");
+	}
+#endif /* Libburn_debug_cd_texT */
+
+	/* Chop from 8 bit text pack to 6 bit subchannel */
+	BURN_ALLOC_MEM(subdata, unsigned char, o->num_text_packs * 24);
+	for (i = 0; i < 18 * o->num_text_packs; i += 3) {
+		si = i / 3 * 4;
+		subdata[si + 0] =  (o->text_packs[i + 0] >> 2) & 0x3f;
+		subdata[si + 1] =  (o->text_packs[i + 0] << 4) & 0x30;
+		subdata[si + 1] |= (o->text_packs[i + 1] >> 4) & 0x0f;
+		subdata[si + 2] =  (o->text_packs[i + 1] << 2) & 0x3c;
+		subdata[si + 2] |= (o->text_packs[i + 2] >> 6) & 0x03;
+		subdata[si + 3] =  (o->text_packs[i + 2] >> 0) & 0x3f;
+	}
+
+	/* Start at Lead-in address of ATIP and write blocks up to -150 */
+	BURN_ALLOC_MEM(buf, struct buffer, 1);
+	write_lba = d->start_lba;
+	for (lba = d->start_lba; lba < -150; lba++) {
+		/* Collect subdata in buf */
+		for (j = 0; j < 4; j++) {
+			memcpy(buf->data + buf->bytes,
+				subdata + sub_cursor * 24, 24);
+			sub_cursor = (sub_cursor + 1) % o->num_text_packs;
+			buf->bytes += 24;
+		}
+		buf->sectors++;
+		sectors++;
+
+		/* When full or last sector : perform WRITE */
+		if (buf->bytes + 96 >= 32768 || lba == -151) {
+
+#ifdef Libburn_debug_cd_texT
+			fprintf(stderr,
+			    "libburn_DEBUG: 6 bit data to be transmitted:\n"); 
+			for (i = 0; i < buf->bytes; i += 24) {
+				fprintf(stderr, "%4d :", i / 24);
+				for (j = 0; j < 24; j++)
+					fprintf(stderr, " %2.2X",
+						buf->data[i + j]);
+				fprintf(stderr, "\n");
+			}
+#endif /* Libburn_debug_cd_texT */
+	
+			err = d->write(d, write_lba, buf);
+			if (err == BE_CANCELLED)
+				{ ret = 0; goto ex; }
+			write_lba += sectors;
+			sectors = buf->sectors = buf->bytes = 0;
+		}
+	}
+	ret = 1;
+ex:;
+	if (self_made_text_packs) {
+		if (o->text_packs != NULL)
+			free(o->text_packs);
+		o->text_packs = NULL;
+		o->num_text_packs = 0;
+	}
+	BURN_FREE_MEM(subdata);
+	BURN_FREE_MEM(buf);
+	d->busy = was_busy;
+	return ret;
+}
+
+
 int burn_write_session(struct burn_write_opts *o, struct burn_session *s)
 {
 	struct burn_drive *d = o->drive;
 	int i, ret;
 
+	if (o->write_type == BURN_WRITE_SAO) {
+		ret = burn_write_leadin_cdtext(o, s, 0);
+		if (ret <= 0)
+			goto ex;
+	}
 	d->rlba = 0;
 	for (i = 0; i < s->tracks; i++) {
 		if (!burn_write_track(o, s, i))
@@ -672,12 +1047,15 @@ ex:;
 
 /* ts A61218 : outsourced from burn_write_track() */
 int burn_disc_init_track_status(struct burn_write_opts *o,
-				struct burn_session *s, int tnum, int sectors)
+				struct burn_session *s, struct burn_track *t,
+				int tnum, int sectors)
 {
 	struct burn_drive *d = o->drive;
 
 	/* Update progress */
+
 	d->progress.start_sector = d->nwa;
+
 	d->progress.sectors = sectors;
 	d->progress.sector = 0;
 
@@ -685,6 +1063,13 @@ int burn_disc_init_track_status(struct burn_write_opts *o,
            by bonfire-app@wanadoo.fr in http://libburn.pykix.org/ticket/58 */
         d->progress.track = tnum;
 
+	/* ts B20113 */
+	d->progress.indices = t->indices;
+	d->progress.index = 0;
+	if (d->progress.indices > 1)
+		if (t->index[0] == 0x7fffffff)
+			d->progress.index = 1;
+
 	/* ts A61102 */
 	d->busy = BURN_DRIVE_WRITING;
 
@@ -712,7 +1097,7 @@ int burn_write_track(struct burn_write_opts *o, struct burn_session *s,
 		if (t->pregap1)
 			d->rlba += 75;
 		if (t->pregap2)
-			d->rlba += 150;
+			d->rlba += t->pregap2_size;
 
 		if (t->pregap1) {
 
@@ -742,13 +1127,20 @@ int burn_write_track(struct burn_write_opts *o, struct burn_session *s,
 					{ ret = 0; goto ex; }
 		}
 		if (t->pregap2)
-			for (i = 0; i < 150; i++)
+			for (i = 0; i < t->pregap2_size; i++)
 				if (!sector_pregap(o, t->entry->point,
 					           t->entry->control, t->mode))
 					{ ret = 0; goto ex; }
+
+		/* ts B20113 : Flush buffer to avoid influence pregap
+		               on track counter */
+		ret = sector_write_buffer(d, NULL, 0);
+		if (ret <= 0)
+			goto ex;
+
 	} else {
 		o->control = t->entry->control;
-		d->send_write_parameters(d, o);
+		d->send_write_parameters(d, s, tnum, o);
 
 		/* ts A61103 */
 		ret = d->get_nwa(d, -1, &lba, &nwa);
@@ -759,9 +1151,11 @@ int burn_write_track(struct burn_write_opts *o, struct burn_session *s,
 		/* <<< */
 		sprintf(msg, 
 	"TAO pre-track %2.2d : get_nwa(%d)=%d, d=%d , demand=%.f , cap=%.f\n",
-	tnum+1, nwa, ret, d->nwa, (double) burn_track_get_sectors(t) * 2048.0,
-	(double) d->media_capacity_remaining);
-		libdax_msgs_submit(libdax_messenger, d->global_index, 0x000002,
+			tnum+1, nwa, ret, d->nwa,
+			(double) burn_track_get_sectors_2(t, 1) * 2048.0,
+			(double) d->media_capacity_remaining);
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+				 0x00000002,
 				LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_ZERO,
 				msg, 0, 0);
 
@@ -781,10 +1175,10 @@ int burn_write_track(struct burn_write_opts *o, struct burn_session *s,
 
 /* user data */
 
-	sectors = burn_track_get_sectors(t);
+	sectors = burn_track_get_sectors_2(t, 1);
 	open_ended = burn_track_is_open_ended(t);
 
-	burn_disc_init_track_status(o, s, tnum, sectors);
+	burn_disc_init_track_status(o, s, t, tnum, sectors);
 
 	/* ts A61030 : this cannot happen. tnum is always < s->tracks */
 	if (tnum == s->tracks)
@@ -827,29 +1221,20 @@ int burn_write_track(struct burn_write_opts *o, struct burn_session *s,
 		d->progress.sector++;
 	}
 
-	if (t->postgap)
-		for (i = 0; i < 150; i++)
-			if (!sector_postgap(o, t->entry->point, t->entry->control,
-				            t->mode))
-				{ ret = 0; goto ex; }
-	i = t->offset;
-	if (o->write_type == BURN_WRITE_SAO) {
-		if (d->buffer->bytes) {
-			int err;
-			err = d->write(d, d->nwa, d->buffer);
-			if (err == BE_CANCELLED)
-				{ ret = 0; goto ex; }
+	/* ts B20113 : Flush buffer to get buffered bytes assigned to the
+	               track counter */
+	ret = sector_write_buffer(d, t, 0);
+	if (ret <= 0)
+		goto ex;
 
-			/* A61101 : probably this is not all payload data */
-			/* A61108 : but audio count is short without this */
-			t->writecount += d->buffer->bytes;
-			t->written_sectors += d->buffer->sectors;
-			d->progress.buffered_bytes += d->buffer->bytes;
-
-			d->nwa += d->buffer->sectors;
-			d->buffer->bytes = 0;
-			d->buffer->sectors = 0;
-		}
+	if (t->postgap && o->write_type != BURN_WRITE_TAO) {
+		for (i = 0; i < t->postgap_size; i++)
+			if (!sector_postgap(o, t->entry->point,
+						 t->entry->control, t->mode))
+				{ ret = 0; goto ex; }
+		ret = sector_write_buffer(d, NULL, 0);
+		if (ret <= 0)
+			goto ex;
 	}
 
 	/* ts A61103 */
@@ -967,6 +1352,35 @@ int burn_disc_init_write_status(struct burn_write_opts *o,
 }
 
 
+static int precheck_write_is_audio(struct burn_disc *disc, int flag)
+{
+	struct burn_session **sessions;
+	int num_sessions, i, j;
+
+	sessions = burn_disc_get_sessions(disc, &num_sessions);
+	for (i = 0; i < num_sessions; i++)
+		for (j = 0; j < sessions[i]->tracks; j++)
+			if (!(sessions[i]->track[j]->mode & BURN_AUDIO))
+				return 0;
+	return 1;
+}
+
+
+static int precheck_disc_has_cdtext(struct burn_disc *disc, int flag)
+{
+	struct burn_session **sessions;
+	int num_sessions, i, ret;
+
+	sessions = burn_disc_get_sessions(disc, &num_sessions);
+	for (i = 0; i < num_sessions; i++) {
+		ret = burn_cdtext_from_session(sessions[i], NULL, NULL, 1);
+		if (ret > 0)
+			return 1;
+	}
+	return 0;
+}
+
+
 /* ts A70219 : API */
 int burn_precheck_write(struct burn_write_opts *o, struct burn_disc *disc,
 				 char reasons[BURN_REASONS_LEN], int silent)
@@ -974,7 +1388,7 @@ int burn_precheck_write(struct burn_write_opts *o, struct burn_disc *disc,
 	enum burn_write_types wt;
 	struct burn_drive *d = o->drive;
 	char *msg = NULL, *reason_pt;
-	int no_media = 0;
+	int no_media = 0, ret, has_cdtext;
 
 	reason_pt= reasons;
 	reasons[0] = 0;
@@ -1001,6 +1415,22 @@ int burn_precheck_write(struct burn_write_opts *o, struct burn_disc *disc,
 	reason_pt= reasons + strlen(reasons);
 	if (d->status == BURN_DISC_UNSUITABLE)
 		goto unsuitable_profile;
+	if (o->num_text_packs > 0) {
+		has_cdtext = 1;
+	} else {
+		has_cdtext = precheck_disc_has_cdtext(disc, 0);
+	}
+	if (has_cdtext > 0) {
+		if (d->current_profile == 0x09 || d->current_profile == 0x0a) {
+			ret = precheck_write_is_audio(disc, 0);
+			if (ret <= 0)
+				strcat(reasons,
+			"CD-TEXT supported only with pure audio CD media, ");
+		} else {
+			strcat(reasons,
+				"CD-TEXT supported only with CD media, ");
+		}
+	}
 	if (d->drive_role == 2 || d->drive_role == 5 ||
 		d->current_profile == 0x1a || d->current_profile == 0x12 ||
 		d->current_profile == 0x43) { 
@@ -1014,6 +1444,14 @@ int burn_precheck_write(struct burn_write_opts *o, struct burn_disc *disc,
 			strcat(reasons, "unsuitable track mode found, ");
 		if (o->start_byte >= 0)
 			strcat(reasons, "write start address not supported, ");
+		if (o->num_text_packs > 0) {
+			if (o->write_type != BURN_WRITE_SAO)
+				strcat(reasons,
+				"CD-TEXT supported only with write type SAO, ");
+			if (d->start_lba == -2000000000)
+				strcat(reasons,
+				"No Lead-in start address known with CD-TEXT, ");
+		}
 	} else if (d->current_profile == 0x13) {
 		/* DVD-RW Restricted Overwrite */
 		if (o->start_byte >= 0 && (o->start_byte % 32768))
@@ -1076,12 +1514,12 @@ int burn_disc_open_track_dvd_minus_r(struct burn_write_opts *o,
 	off_t size;
 
 	BURN_ALLOC_MEM(msg, char, 160);
-	d->send_write_parameters(d, o);
+	d->send_write_parameters(d, NULL, -1, o);
 	ret = d->get_nwa(d, -1, &lba, &nwa);
 	sprintf(msg, 
 		"DVD pre-track %2.2d : get_nwa(%d), ret= %d , d->nwa= %d",
 		tnum+1, nwa, ret, d->nwa);
-	libdax_msgs_submit(libdax_messenger, d->global_index, 0x000002,
+	libdax_msgs_submit(libdax_messenger, d->global_index, 0x00000002,
 			LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_ZERO, msg,0,0);
 	if (nwa > d->nwa)
 		d->nwa = nwa;
@@ -1102,7 +1540,7 @@ int burn_disc_open_track_dvd_minus_r(struct burn_write_opts *o,
 #endif
 
 	if (o->write_type == BURN_WRITE_SAO) { /* DAO */
-		size = ((off_t) burn_track_get_sectors(s->track[tnum]))
+		size = ((off_t) burn_track_get_sectors_2(s->track[tnum], 1))
 			* (off_t) 2048;
 
  		/* Eventually round track size up to write chunk */
@@ -1141,7 +1579,7 @@ int burn_disc_open_track_dvd_plus_r(struct burn_write_opts *o,
 	sprintf(msg, 
 		"DVD+R pre-track %2.2d : get_nwa(%d), ret= %d , d->nwa= %d",
 		tnum+1, nwa, ret, d->nwa);
-	libdax_msgs_submit(libdax_messenger, d->global_index, 0x000002,
+	libdax_msgs_submit(libdax_messenger, d->global_index, 0x00000002,
 			LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_ZERO, msg,0,0);
 	if (nwa > d->nwa)
 		d->nwa = nwa;
@@ -1151,7 +1589,7 @@ int burn_disc_open_track_dvd_plus_r(struct burn_write_opts *o,
 	if (o->write_type == BURN_WRITE_SAO &&
 	    ! burn_track_is_open_ended(s->track[tnum])) {
  		/* Round track size up to write chunk size and reserve track */
-		size = ((off_t) burn_track_get_sectors(s->track[tnum]))
+		size = ((off_t) burn_track_get_sectors_2(s->track[tnum], 1))
 			* (off_t) 2048;
 		/* o->obs should be 32k or 64k already. But 32k alignment
 		  was once performed in d->reserve_track() */
@@ -1425,12 +1863,12 @@ int burn_dvd_write_track(struct burn_write_opts *o,
 			goto ex;
 	}
 
-	sectors = burn_track_get_sectors(t);
+	sectors = burn_track_get_sectors_2(t, 1);
 	open_ended = burn_track_is_open_ended(t);
 
 	/* (offset padding is done within sector_data()) */
 
-	burn_disc_init_track_status(o, s, tnum, sectors);
+	burn_disc_init_track_status(o, s, t, tnum, sectors);
 	for (i = 0; open_ended || i < sectors; i++) {
 
 		/* From time to time inquire drive buffer */
@@ -1747,7 +2185,7 @@ int burn_disc_setup_dvd_minus_rw(struct burn_write_opts *o,
 	5.4.14 finally states that profile 0013h includes feature
 	002Ch rather than 0026h.
 		
-		d->send_write_parameters(d, o);
+		d->send_write_parameters(d, NULL, -1, o);
 	*/
 
 	d->busy = BURN_DRIVE_FORMATTING;
@@ -2268,8 +2706,8 @@ int burn_stdio_write_track(struct burn_write_opts *o, struct burn_session *s,
 
 	BURN_ALLOC_MEM(buf, char, bufsize);
 
-	sectors = burn_track_get_sectors(t);
-	burn_disc_init_track_status(o, s, tnum, sectors);
+	sectors = burn_track_get_sectors_2(t, 1);
+	burn_disc_init_track_status(o, s, t, tnum, sectors);
 	open_ended = burn_track_is_open_ended(t);
 
 	t->end_on_premature_eoi = (o->write_type == BURN_WRITE_TAO);
@@ -2386,6 +2824,7 @@ void burn_disc_write_sync(struct burn_write_opts *o, struct burn_disc *disc)
 	struct cue_sheet *sheet;
 	struct burn_drive *d = o->drive;
 	struct buffer *buffer_mem = o->drive->buffer;
+	struct burn_session *s;
 	struct burn_track *lt, *t;
 	int first = 1, i, ret, lba, nwa = 0, multi_mem;
 	off_t default_size;
@@ -2498,8 +2937,11 @@ return crap.  so we send the command, then ignore the result.
 	if (o->write_type == BURN_WRITE_TAO) {
 		nwa = 0; /* get_nwa() will be called in burn_track() */
 	} else {
-
-		d->send_write_parameters(d, o);
+		if (disc->sessions > 0)
+			s = disc->session[0];
+		else
+			s = NULL;
+		d->send_write_parameters(d, s, -1, o);
 
 		ret = d->get_nwa(d, -1, &lba, &nwa);
 		sprintf(msg,
@@ -2524,14 +2966,23 @@ return crap.  so we send the command, then ignore the result.
 
 		/* ts A61009 */
 		if (sheet == NULL)
-			goto fail;
+			goto fail_wo_sync;
 
-/*		print_cue(sheet);*/
+#ifdef Libburn_write_with_function_print_cuE
+		print_cue(sheet);
+		/* goto fail_wo_sync; */
+#endif /* Libburn_write_with_function_print_cuE */
+
+		ret = 1;
 		if (o->write_type == BURN_WRITE_SAO)
-			d->send_cue_sheet(d, sheet);
+			ret = d->send_cue_sheet(d, sheet);
 		if (sheet->data != NULL)
 			free(sheet->data);
 		free(sheet);
+		if (ret <= 0)
+			goto fail_wo_sync;
+
+		/* --- From here on, final sync is needed. --- */
 
 		if (o->write_type == BURN_WRITE_RAW) {
 			if (!burn_write_leadin(o, disc->session[i], first))
@@ -2555,7 +3006,7 @@ return crap.  so we send the command, then ignore the result.
 						sprintf(msg, 
 				"SAO appendable d->nwa= %d\n", d->nwa);
 						libdax_msgs_submit(
-				libdax_messenger, d->global_index, 0x000002,
+				libdax_messenger, d->global_index, 0x00000002,
 				LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_ZERO,
 				msg, 0, 0);
 
@@ -2812,7 +3263,7 @@ int burn_disc_close_damaged(struct burn_write_opts *o, int flag)
 		o->write_type = BURN_WRITE_TAO; /* no action without TAO */
 
 		/* Send mode page 5 */;
-		d->send_write_parameters(d, o);
+		d->send_write_parameters(d, NULL, -1, o);
 
 		ret = burn_write_close_session(o);
 		if (ret <= 0)
@@ -2823,7 +3274,7 @@ int burn_disc_close_damaged(struct burn_write_opts *o, int flag)
 		o->write_type = BURN_WRITE_TAO; /* no action without TAO */
 
 		/* Send mode page 5 */;
-		d->send_write_parameters(d, o);
+		d->send_write_parameters(d, NULL, -1, o);
 
 		ret = burn_disc_close_track_dvd_minus_r(o, 0);
 		if (ret <= 0)
